/* -*-mode:C;c-basic-offset:8-*-
 *
 *      Copyright (C) 1993-1994 Bas Laarhoven.
 *
 * some changes by Claus-Justus Heine, 1994, 1995, 1996
 * implemented dynamic memory allocation, see zftape-buffers.c
 *
 

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2, or (at your option)
any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; see the file COPYING.  If not, write to
the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.

 $Source: /homes/cvs/zftape/kernel-interface.c,v $
 $Author: claus $
 *
 $Revision: 1.26 $
 $Date: 1996/08/31 00:48:52 $
 $State: Exp $
 *
 *      This file contains the code that interfaces the kernel
 *      for the QIC-40/80 floppy-tape driver for Linux.
 */

static char RCSid[] =
"$Id: kernel-interface.c,v 1.26 1996/08/31 00:48:52 claus Exp $";

#include "ftape.h"
#include <linux/module.h>
#include <linux/errno.h>
#include <linux/fs.h>

#include "fdc-io.h"             /* for ftape_unit */
#include "kernel-interface.h"
#include "zftape-buffers.h"
#include "ftape-ctl.h"
#include "ftape-read.h"
#include "ftape-write.h"
#include "ftape-io.h"

/*      Global vars.
 */
/*
 *  1.3.38 linux/module.h defines kernel_version itself, unless
 *  we define __NO_VERSION__
 */
#ifdef __NO_VERSION__
char kernel_version[] = UTS_RELEASE;
#endif

int busy_flag = 0;
/*      Local vars.
 */
static int old_sigmask;

static int  ftape_open ( struct inode* ino, struct file* filep);
static void ftape_close( struct inode* ino, struct file* filep);
static int  ftape_ioctl( struct inode* ino, struct file* filep, 
			 unsigned int command, unsigned long arg);
static int  ftape_read ( struct inode* ino, struct file* fp, char* buff,
			 int req_len);
#if LINUX_VERSION_CODE >= KERNEL_VER(1,3,0)
static int  ftape_write( struct inode* ino, struct file* fp, const char* buff,
			 int req_len);
#else
static int  ftape_write( struct inode* ino, struct file* fp, char* buff,
			 int req_len);
#endif
static int  ftape_lseek( struct inode* ino, struct file* filep,
			 off_t offset, int origin);
#if 0
static int  ftape_select( void);
static int  ftape_mmap  ( int dev, unsigned off, int prot);
#else
# define ftape_select NULL
# define ftape_mmap NULL
#endif

static struct file_operations ftape_cdev = {
	ftape_lseek,		/* lseek */
	ftape_read, 		/* read */
	ftape_write,		/* write */
	NULL,			/* readdir */
	ftape_select, 		/* select */
	ftape_ioctl,		/* ioctl */
	ftape_mmap, 		/* mmap */
	ftape_open,		/* open */
	ftape_close,		/* release */
	NULL,			/* fsync */
};

/*      Called by modules package when installing the driver
 */

int
init_module( void) {
	TRACE_FUN( 5, "init_module");

	printk( KERN_INFO "\n"
	        KERN_INFO "zftape v1.06 14/03/96\n"
	        KERN_INFO "(c) 1994-1996 Claus-Justus Heine "
		          "(Heine@axpmgr.physik.rwth-aachen.de)\n"
		KERN_INFO "Support for real blocks of constant length, "
	                  "dynamic memory allocation\n"
		KERN_INFO "and compression (lzrw3-algorithm). "
		          "Compiled for kernel version %s"
#ifdef MODVERSIONS
	                  "\n" KERN_INFO "with versioned symbols"
#endif
	                  ".\n", kernel_version);
	printk( KERN_INFO "This floppy tape driver was build on top of:\n");
	printk(KERN_INFO "ftape v2.09 960616\n");
        if(tracing>=3) {
		printk(KERN_INFO "(c) 1993-1995 Bas Laarhoven (bas@vimec.nl)\n"
		       KERN_INFO "(c) 1995-1996 Kai Harrekilde-Petersen (khp@dolphinics.no)\n"
		       KERN_INFO "QIC-117 driver for QIC-40/80/3010/3020 tape drives\n");
	}
	TRACEx1( 3, "init_module @ 0x%p", init_module);
	if (zftape_init_mem() < 0) {
		TRACE( 1, "Error: memory initialization failed");
		TRACE_EXIT;
		return -ENOMEM;
	}
	TRACE( 4, "\ninstalling QIC-117 ftape driver...");
	if (register_chrdev( QIC117_TAPE_MAJOR, "zft", &ftape_cdev)) {
		TRACE( 1, "register_chrdev failed");
		TRACE_EXIT;
		return -EIO;
	}
	busy_flag = 0;
	ftape_unit = -1;
	ftape_failure = 1;            /* inhibit any operation but open */
	udelay_calibrate();           /* must be before fdc_wait_calibrate ! */
	fdc_wait_calibrate();
	TRACE_EXIT;
	register_symtab(0);           /* remove global ftape symbols */
	return 0;
}

/*      Called by modules package when removing the driver
 */
void
cleanup_module( void) {
	TRACE_FUN( 5, "cleanup_module");
	
	if (unregister_chrdev( QIC117_TAPE_MAJOR, "zft") != 0) {
		TRACE( 2, "failed");
	} else {
		TRACE( 4, "successful");
	}
	zftape_uninit_mem();
	TRACE_EXIT;
}

/*      Open ftape device
 */
static int
ftape_open( struct inode* ino, struct file* filep)
{
	TRACE_FUN( 5, "ftape_open");
	int result;

	MOD_INC_USE_COUNT;            /* lock module in memory */
	
	TRACEi( 5, "called for minor", MINOR( ino->i_rdev));
	if (busy_flag) {
		TRACE( 1, "failed: already busy");
		MOD_DEC_USE_COUNT;          /* unlock module in memory */
		TRACE_EXIT;
		return -EBUSY;
	}
#ifdef HAPPY_HACKING 
	if ( MOD_IN_USE ) {
		mod_use_count_ =  0;
	}
#endif
	if ( (MINOR( ino->i_rdev) & ~ZFTAPE_MINOR_MASK) != 0 ) {
		TRACE( 1, "failed: illegal unit nr");
		MOD_DEC_USE_COUNT;          /* unlock module in memory */
		TRACE_EXIT;
		return -ENXIO;
	}
	if ( zftape_get_tape_buffers ( ino->i_rdev ) < MIN_DMA_BUFFERS ) {
		TRACE( 1, "failed: unsufficient memory");
		MOD_DEC_USE_COUNT;
		TRACE_EXIT;
		return -ENOMEM;
	}
	if (ftape_unit == -1 || 
	    FTAPE_UNIT != (MINOR( ino->i_rdev) & ZFTAPE_MINOR_SEL_MASK)) {
		/*  Other selection than last time
		 */
		ftape_init_driver();
	}
	ftape_failure = 0;            /* allow tape operations */
	/* 
	 *  hack for kerneld
	 */
#if defined(CONFIG_KERNELD) && defined(KERNELD_PROTECT)
	if ( (ftape_unit != -1) && (ftape_unit & FTAPE_NO_REWIND) ) {
		MOD_DEC_USE_COUNT;
	}
#endif
	ftape_unit = MINOR( ino->i_rdev);
	old_sigmask = current->blocked;
	current->blocked = _BLOCK_ALL;
	fdc_save_drive_specs(); /* save Drive Specification regs 
				 * on i82078-1's 
				 */
	result = _ftape_open();
	if (result < 0) {
		TRACE( 1, "_ftape_open failed");
		zftape_unget_tape_buffers(); /* release the dma buffer  */
		fdc_restore_drive_specs();  /* restore original values */
		current->blocked = old_sigmask; /* restore mask */
		MOD_DEC_USE_COUNT;              /* unlock module in memory */
		TRACE_EXIT;
		return result;
	} else {
		busy_flag = 1;
		/*  Mask signals that will disturb proper operation of the
		 *  program that is calling.
		 */
		current->blocked = old_sigmask | _DO_BLOCK;
		TRACE_EXIT;
		return 0;
	}
}

/*      Close ftape device
 */
static void
ftape_close( struct inode* ino, struct file* filep)
{
	TRACE_FUN( 5, "ftape_close");
	int result;
	
	if (!busy_flag || MINOR( ino->i_rdev) != ftape_unit) {
		TRACE( 1, "failed: not busy or wrong unit");
		TRACE_EXIT;
		return;                     /* keep busy_flag !(?) */
	}
	current->blocked = _BLOCK_ALL;
	result = _ftape_close();
	if (result < 0) {
		TRACE( 1, "_ftape_close failed");
	}
	fdc_restore_drive_specs();  /* restore original values */
	zftape_unget_tape_buffers(); /* release the dma buffer  */
	ftape_failure = 1;          /* inhibit any operation but open */
	busy_flag = 0;
	current->blocked = old_sigmask; /* restore before open state */
	TRACE_EXIT;
#if defined(CONFIG_KERNELD) && defined(KERNELD_PROTECT)
	/*
	 *  this should enable use with the kerneld daemon.
	 */
	if (!(ftape_unit & FTAPE_NO_REWIND)) {
		MOD_DEC_USE_COUNT;            /* unlock module in memory */
	}
#else
	MOD_DEC_USE_COUNT;            /* unlock module in memory */
#endif
}

/*      Ioctl for ftape device
 */
static int
ftape_ioctl( struct inode* ino, struct file* filep, 
            unsigned int command, unsigned long arg)
{
	TRACE_FUN( 5, "ftape_ioctl");
	int result = -EIO;
	int old_sigmask;
	
	if (!busy_flag || MINOR( ino->i_rdev) != ftape_unit || ftape_failure) {
		TRACE( 1, "failed: not busy, failure or wrong unit");
		TRACEx4(1,"ftape_failure: %d, busy_flag: %d, i_rdev: %08x, "
			  "ftape_unit: %08x", 
			ftape_failure, busy_flag, MINOR(ino->i_rdev), 
			ftape_unit );
		TRACE_EXIT;
		return -EIO;
	}
	old_sigmask = current->blocked; /* save mask */
	current->blocked = _BLOCK_ALL;
	/* This will work as long as sizeof( void*) == sizeof( long)
	 */
	result = _ftape_ioctl( command, (void*) arg);
	current->blocked = old_sigmask; /* restore mask */
	TRACE_EXIT;
	return result;
}

/*      Read from tape device
 */
static int
ftape_read( struct inode* ino, struct file* fp, char* buff, int req_len)
{
	TRACE_FUN( 5, "ftape_read");
	int result = -EIO;
	int old_sigmask;

	TRACEi( 5, "called with count:", req_len);
	if (!busy_flag || MINOR( ino->i_rdev) != ftape_unit || ftape_failure) {
		TRACE( 1, "failed: not busy, failure or wrong unit");
		TRACE_EXIT;
		return -EIO;
	}
	old_sigmask = current->blocked; /* save mask */
	current->blocked = _BLOCK_ALL;
	result = _ftape_read( buff, req_len);
	TRACEi( 7, "return with count:", result);
	current->blocked = old_sigmask; /* restore mask */
	TRACE_EXIT;
	return result;
}

/*      Write to tape device
 */
static int
#if LINUX_VERSION_CODE >= KERNEL_VER(1,3,0)
ftape_write( struct inode* ino, struct file* fp, const char* buff, int req_len)
#else
ftape_write( struct inode* ino, struct file* fp, char* buff, int req_len)
#endif
{
	TRACE_FUN( 5, "ftape_write");
	int result = -EIO;
	int old_sigmask;
	
	TRACEi( 7, "called with count:", req_len);
	if (!busy_flag || MINOR( ino->i_rdev) != ftape_unit || ftape_failure) {
		TRACE( 1, "failed: not busy, failure or wrong unit");
		TRACE_EXIT;
		return -EIO;
	}
	old_sigmask = current->blocked; /* save mask */
	current->blocked = _BLOCK_ALL;
	result = _ftape_write( buff, req_len);
	TRACEi( 7, "return with count:", result);
	current->blocked = old_sigmask; /* restore mask */
	TRACE_EXIT;
	return result;
}

/*      Seek tape device - not implemented !
 */
static int
ftape_lseek( struct inode* ino, struct file* filep, off_t offset, int origin)
{
	return -ESPIPE;
}


